#include <SPI.h>
#include <TFT.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>

// pin connections for the tft screen
#define PIN_TFT_CS 10
#define PIN_DC 9
#define PIN_RST 8

#define NULL_TILE   0
#define LEFT    0
#define RIGHT   1
#define UP      2
#define DOWN    3
#define LOST    4
#define WON     8
#define SIZE    4

// grid related information
#define SIZE 4
#define CELL_WIDTH 36.9
#define CELL_HEIGHT 29
#define CELL_MARGIN 4
#define CELL_RADIUS 5
#define COLOR_GRID_BORDER 0x0000
#define COLOR_TILE_NULL_TILE 0xC618
#define COLOR_TILE_FILLED 0xFFFF

TFT TFTscreen = TFT(PIN_TFT_CS, PIN_DC, PIN_RST);

// tile's colors
#define COLOR_BACKGROUND 0x0000
#define COLOR_NULL_TILE_TILE 0xC618 
#define COLOR_TILE_2    0xFC40  // orange - needs white
#define COLOR_TILE_4    0xB1A0  // dark red - needs white 
#define COLOR_TILE_8    0x001F  // blue - needs white
#define COLOR_TILE_16   0x07E0 // green - needs black
#define COLOR_TILE_32   0x07FF  // light blue - needs black
#define COLOR_TILE_64   0xFFE0  // yellow - needs black
#define COLOR_TILE_128  0xFD20  // orange/brown - needs white
#define COLOR_TILE_256  0xF93F  // purple - needs black
#define COLOR_TILE_512  0xAFE5  // light green - needs black
#define COLOR_TILE_1024 0xA254  // dark blue - needs white
#define COLOR_TILE_2048 0xFF00  // Gold - doesnt matter

// song notes
#define NOTE_C4  261
#define NOTE_CS4 277
#define NOTE_D4  293
#define NOTE_DS4 311
#define NOTE_E4  329
#define NOTE_F4  349
#define NOTE_FS4 369
#define NOTE_G4  392
#define NOTE_GS4 415
#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  493
#define NOTE_C5  523
#define NOTE_G3 196
#define NOTE_REST 0

// used for checking if the player reached x for the first time
int check128 = 0;
int check256 = 0;
int check512 = 0;
int check1024 = 0;

// button pins
const int resetButton = A5;
const int downButton = 7;
const int rightButton = 5;
const int upButton = 3;
const int leftButton = 2;
const int buzzerPin = 6;

// button states
volatile bool buttonLeftPressed = false;
volatile bool buttonUpPressed = false;
volatile int rightButtonState = LOW;
volatile int lastRightButtonState = LOW;
volatile int downButtonState = LOW;
volatile int lastDownButtonState = LOW;
volatile int resetButtonState = 0;

// variables used for the so-called song
int melody[] = {
  NOTE_C4, 250, // C4 for 250 milliseconds
  NOTE_E4, 250, // E4 for 250 milliseconds
  NOTE_G4, 250, // G4 for 250 milliseconds
  NOTE_C5, 500, // C5 for 500 milliseconds
  NOTE_G4, 250, // G4 for 250 milliseconds
  NOTE_E4, 250, // E4 for 250 milliseconds
  NOTE_C4, 500, // C4 for 500 milliseconds
};
int noteDurations[] = {
  4, 4, 4, 2, 4, 4, 2
};
int pauseDuration = 100;

int lostMelody[] = {
  NOTE_E4, NOTE_E4, NOTE_REST, NOTE_E4, NOTE_REST, NOTE_C4, NOTE_E4, NOTE_REST,
  NOTE_G4, NOTE_REST, NOTE_G3, NOTE_REST, NOTE_G3, NOTE_REST, NOTE_C4, NOTE_REST,
  NOTE_D4, NOTE_REST, NOTE_REST, NOTE_REST, NOTE_REST, NOTE_REST, NOTE_REST, NOTE_REST
};

int lostNoteDurations[] = {
  500, 500, 500, 250, 250, 500, 250, 250,
  500, 250, 250, 500, 250, 250, 750, 250,
  500, 250, 250, 250, 250, 250, 250, 250
};

int lostPauseDuration = 100; // pause duration between notes in milliseconds
int lostNoteIndex = 0; // 

uint16_t grid[SIZE][SIZE];

void setupGrid(void)
{
    memset(grid, 0, SIZE * SIZE * sizeof(uint16_t));
    newTile();
}

void newTile()
{
    uint16_t new_value, i, j;
    uint16_t random_seed = rand() % 100;
    if (random_seed > 60)
        new_value = 4;
    else
        new_value = 2;

    random_seed = rand() % 16;
    i = random_seed / SIZE;
    j = random_seed % SIZE;
    
    grid[i][j] = new_value;
}

void swap_tiles(uint16_t *tile1, uint16_t *tile2)
{
uint16_t temp = *tile1;
*tile1 = *tile2;
*tile2 = temp;
}

void merge_tiles(uint16_t *tile1, uint16_t *tile2)
{
*tile1 = *tile1 * 2;
*tile2 = NULL_TILE;
}

void move_left(uint16_t i, uint16_t speed[SIZE])
{
  uint16_t j, c;
  uint16_t merge_flag = 0;
  
  memset(speed, 0, SIZE);
  
  for (j = 1; j < SIZE; ++j)
  {
        if (grid[i][j] !=  NULL_TILE)
        {
            c = j;
            while (c > 0 && grid[i][c - 1] == NULL_TILE)
            {
                swap_tiles(&grid[i][c - 1], &grid[i][c]);
                --c;
            }
            if (c > 0 && grid[i][c - 1] == grid[i][c] && !merge_flag)
            {
                merge_tiles(&grid[i][c - 1], &grid[i][c]);
                merge_flag = 1;
            }
        }
}
}

// grid[i][0] <-> grid[i][3], grid[i][1] <-> m[i][2]
void reflect(uint16_t m[SIZE][SIZE])
{
  uint16_t i, j, temp;
  
  for (i = 0; i < SIZE; i++) {
      for (j = 0; j < SIZE / 2; j++) {
          temp = m[i][j];
          m[i][j] = m[i][SIZE - 1 - j];
          m[i][SIZE - 1 - j] = temp;
      }
  }
}

// used for up/down
void rotate(uint16_t mat[SIZE][SIZE], uint16_t type)
{
    uint16_t i, j;
    uint16_t newGrid[SIZE][SIZE];
    
    if (type == LEFT) {
        // clock-wise
        for (i = 0; i < SIZE; ++i)
            for (j = 0; j < SIZE; ++j)
                newGrid[i][j] = mat[j][SIZE - i - 1];
    } else {
        // anti cock-wise
        for (i = 0; i < SIZE; ++i)
            for (j = 0; j < SIZE; ++j)
                newGrid[i][j] = mat[SIZE - j - 1][i];
    }
    
    memcpy(mat, newGrid, SIZE * SIZE * sizeof(uint16_t));
}

void move(uint16_t type, uint16_t speeds[SIZE][SIZE])
{
    switch (type)
    {
        case RIGHT:
            reflect(grid);
            break;
        case UP:
            rotate(grid, LEFT);
            break;
        case DOWN:
            rotate(grid, RIGHT);
            break;
    }
    
    memset(speeds, 0, SIZE * SIZE * sizeof(uint16_t));  
    uint16_t i;  
    for (i = 0; i < SIZE; ++i)
        move_left(i, speeds[i]);
        
    switch (type)
    {
        case RIGHT:
            reflect(grid);
            break;
        case UP:
            rotate(grid, RIGHT);
            break;
        case DOWN:
            rotate(grid, LEFT);
            break;
    }
}

int areMatricesEqual(uint16_t matrix1[SIZE][SIZE], uint16_t matrix2[SIZE][SIZE])
{
    for (uint16_t i = 0; i < SIZE; ++i){
        for (uint16_t j = 0; j < SIZE; ++j){
            if (matrix1[i][j] != matrix2[i][j])
              return 1;
        }
    }
    return 0;
}

uint16_t test(uint16_t type, uint16_t speeds[SIZE][SIZE])
{
    uint16_t temp[SIZE][SIZE];
    memcpy(temp, grid, SIZE * SIZE * sizeof(uint16_t));
    move(type, speeds);
    uint16_t res = areMatricesEqual(temp, grid);
    memcpy(grid, temp, SIZE * SIZE * sizeof(uint16_t));
    return res;
}

uint16_t check(void)
{
    uint16_t i, j;
    uint16_t testt[SIZE][SIZE];
    
    if (!(test(LEFT, testt) || test(RIGHT, testt) || test(UP, testt) || test(DOWN, testt)))
        return LOST;
    
    for (i = 0; i < SIZE; ++i)
        for (j = 0; j < SIZE; ++j)
            if (grid[i][j] == 2048)
                return WON;
                
    return 200;
}

void playWonMelody() {
  for (int i = 0; i < sizeof(melody) / sizeof(melody[0]); i += 2) {
    int noteFrequency = melody[i];
    int noteDuration = noteDurations[i / 2];
    int period = 1000000 / noteFrequency;

    analogWrite(buzzerPin, HIGH);
    delayMicroseconds(period * 0.9); // Play the note for 90% of the duration

    analogWrite(buzzerPin, LOW);
    delay(noteDuration * 1.1); // Pause for 10% of the duration

    delay(pauseDuration); // Pause between notes
  }
}

void playLostMelody() {
  for (int i = 0; i < sizeof(lostMelody) / sizeof(lostMelody[0]); i += 2) {
    int noteFrequency = lostMelody[i];
    int noteDuration = lostNoteDurations[i / 2];
    int period = 1000000 / noteFrequency;

    analogWrite(buzzerPin, HIGH);
    delayMicroseconds(period * 0.9); // Play the note for 90% of the duration

    analogWrite(buzzerPin, LOW);
    delay(noteDuration * 1.1); // Pause for 10% of the duration

    delay(pauseDuration); // Pause between notes
  }
}

void handleLeftInterrupt() {
  buttonLeftPressed = true;
}

void handleUpInterrupt() {
  buttonUpPressed = true;
}

bool isValueInGrid(int value) {
  for (int i = 0; i < SIZE; i++) {
    for (int j = 0; j < SIZE; j++) {
      if (grid[i][j] == value) {
        return true;
      }
    }
  }
  return false;
}

void drawGrid() {
  for (int i = 0; i < SIZE; i++) {
    for (int j = 0; j < SIZE; j++) {
      // calculate the x and y of the current tile
      int x = j * (CELL_WIDTH + CELL_MARGIN);
      int y = i * (CELL_HEIGHT + CELL_MARGIN);

      uint16_t tileColor;
      switch (grid[i][j]) {
        case 0:
          tileColor = COLOR_NULL_TILE_TILE;
          break;
        case 2:
          tileColor = COLOR_TILE_2;
          break;
        case 4:
          tileColor = COLOR_TILE_4;
          break;
        case 8:
          tileColor = COLOR_TILE_8;
          break;
        case 16:
          tileColor = COLOR_TILE_16;
          break;
        case 32:
          tileColor = COLOR_TILE_32;
          break;
        case 64:
          tileColor = COLOR_TILE_64;
          break;
        case 128:
          tileColor = COLOR_TILE_128;
          break;
        case 256:
          tileColor = COLOR_TILE_256;
          break;
        case 512:
          tileColor = COLOR_TILE_512;
          break;
        case 1024:
          tileColor = COLOR_TILE_1024;
          break;
        case 2048:
          tileColor = COLOR_TILE_2048;
          break;
        default:
          tileColor = COLOR_BACKGROUND;
          break;
      }
      TFTscreen.fillRoundRect(x, y, CELL_WIDTH, CELL_HEIGHT, CELL_RADIUS, tileColor);
      TFTscreen.drawRoundRect(x, y, CELL_WIDTH, CELL_HEIGHT, CELL_RADIUS, COLOR_GRID_BORDER);
      
      if (grid[i][j] != 0) {
        // Set the font color based on the tile color for better contrast
        uint16_t fontColor = (tileColor == COLOR_NULL_TILE_TILE || 
                              tileColor == COLOR_TILE_2 || 
                              tileColor == COLOR_TILE_4 ||
                              tileColor == COLOR_TILE_8 ||
                              tileColor == COLOR_TILE_128 ||
                              tileColor == COLOR_TILE_1024) ? 0xFFFF : 0x0000;
        TFTscreen.setTextColor(fontColor);
        TFTscreen.setTextSize(1);
        
        // Calculate the position to center the text within the tile
        // the values have been calculated empirically
        int resize;
        if (grid[i][j] < 10)
          resize = 5/2;
         else if (grid[i][j] < 65)
          resize = 10;
         else if (grid[i][j] < 513)
          resize = 16;
         else if (grid[i][j] < 2049)
          resize = 24;
        TFTscreen.setCursor(x + (CELL_WIDTH - resize) / 2, y + CELL_HEIGHT / 2 - 5/2);
        TFTscreen.print(grid[i][j]);
      }
    }
  }
}

void setup() {
  
  Serial.begin(9600);

  // need the srand in order to have a random matrix
  srand(analogRead(A0));
  setupGrid();
  
  pinMode(buzzerPin, OUTPUT);
  
  pinMode(resetButton, INPUT_PULLUP);
  pinMode(upButton, INPUT_PULLUP);
  pinMode(rightButton, INPUT_PULLUP);
  pinMode(downButton, INPUT_PULLUP);
  pinMode(leftButton, INPUT_PULLUP);
  
  digitalWrite(resetButton, HIGH);

  attachInterrupt(digitalPinToInterrupt(leftButton), handleLeftInterrupt, FALLING);
  attachInterrupt(digitalPinToInterrupt(upButton), handleUpInterrupt, FALLING);
  
  TFTscreen.initR(INITR_BLACKTAB);
  // clockwise 90 degrees rotation bcs bs orientation 
  // spent 1h searching for this function :)
  TFTscreen.setRotation(1);
  TFTscreen.background(255, 255, 255);

  // activate the interrupts
  sei();
}
// Main loop function
void printWonScreen() {
  TFTscreen.fillScreen(0xFFFF);
  TFTscreen.setTextSize(5);
  TFTscreen.setTextColor(COLOR_TILE_2048);
  TFTscreen.setCursor(40, 47);
  TFTscreen.println("WON");
}

void printLostScreen() {
  TFTscreen.fillScreen(0xFFFF);
  TFTscreen.setTextSize(5);
  TFTscreen.setTextColor(COLOR_TILE_256);
  TFTscreen.setCursor(25, 47);
  TFTscreen.println("LOST");
}

void printValue(int value) {
  if (value != 1024) {
    TFTscreen.fillScreen(0xFFFF);
    TFTscreen.setTextSize(5);
    TFTscreen.setTextColor(COLOR_TILE_2048);
    TFTscreen.setCursor(40, 47);
  } else {
    TFTscreen.fillScreen(0xFFFF);
    TFTscreen.setTextSize(5);
    TFTscreen.setTextColor(COLOR_TILE_256);
    TFTscreen.setCursor(25, 47);
  }
    TFTscreen.println(value);
}

void loop() {
  int res = check();
  if (res == LOST) {
    Serial.println("LOST");
    printLostScreen();
    while (analogRead(resetButton) > 100) {
      playLostMelody();
    }
    check128 = 0;
    check256 = 0;
    check512 = 0;
    check1024 = 0;
    setupGrid();
    TFTscreen.background(255, 255, 255);
  } else if (res == WON) {
      Serial.println("WON");
      Serial.println("LOST");
      printWonScreen();
      while (analogRead(resetButton) > 100) {
        playWonMelody();
      }
      check128 = 0;
      check256 = 0;
      check512 = 0;
      check1024 = 0;
      setupGrid();
      TFTscreen.background(255, 255, 255);
  } else {
    drawGrid();
  }
  
  resetButtonState = analogRead(resetButton);
  if (resetButtonState < 100) {
    Serial.println("Reset Button Pressed");
    check128 = 0;
    check256 = 0;
    check512 = 0;
    check1024 = 0;
    setupGrid();
    newTile();
    TFTscreen.background(255, 255, 255);
  }

  rightButtonState = digitalRead(rightButton);
  if (rightButtonState != lastRightButtonState) {
    if (rightButtonState == HIGH) {
      uint16_t speeds[SIZE][SIZE];  // Create an NULL_TILE speeds array
      if (test(RIGHT  , speeds)) {
        move(RIGHT, speeds);
        newTile();
        Serial.println("Right Button pressed");
      }
    }
  }
  lastRightButtonState = rightButtonState;

  downButtonState = digitalRead(downButton);
  if (downButtonState != lastDownButtonState) {
    if (downButtonState == HIGH) {
      uint16_t speeds[SIZE][SIZE];  // Create an NULL_TILE speeds array
      if (test(DOWN, speeds)) {
        move(DOWN, speeds);
        newTile();
        Serial.println("Down Button pressed");
      }
    }
  }
  lastDownButtonState = downButtonState;
  
  if (buttonLeftPressed) {
      uint16_t speeds[SIZE][SIZE];  // Create an NULL_TILE speeds array
      if (test(LEFT, speeds)) {
        move(LEFT, speeds);
        newTile();
        Serial.println("Left Button pressed!");
      }
    buttonLeftPressed = false;
  }
  
  if (buttonUpPressed) {
    uint16_t speeds[SIZE][SIZE];  // Create an NULL_TILE speeds array
    if (test(UP, speeds)) {
      move(UP, speeds);
      newTile();
      Serial.println("Up Button pressed!");
    }
    buttonUpPressed = false;
  }
  if ((check128 == 0) && isValueInGrid(128)){
    Serial.println("here");
    unsigned long startTime = millis();
    Serial.println(millis());
    Serial.println(startTime);
    while (millis() - startTime < 5000) {
      playWonMelody();
      printValue(128);
    }
    check128 = 1;
    TFTscreen.background(255, 255, 255);
  } else if ((check256 == 0) && isValueInGrid(256)){
    unsigned long startTime = millis();
    while (millis() - startTime < 5000) {
      playWonMelody();
      printValue(256);
    }
    check256 = 1;
    TFTscreen.background(255, 255, 255);
  } else if ((check512 == 0) && isValueInGrid(512)){
    unsigned long startTime = millis();
    while (millis() - startTime < 5000) {
      playWonMelody();
      printValue(512);
    }
    check512 = 1;
    TFTscreen.background(255, 255, 255);
  } else if ((check1024 == 0) && isValueInGrid(1024)){
    unsigned long startTime = millis();
    while (millis() - startTime < 5000) {
      playWonMelody();
      printValue(1024);
    }
    check1024 = 1;
    TFTscreen.background(255, 255, 255);
  }
}
